/***********************************************************************************
*
*                                                                        2010-08-12
* 
* This program was written for the programming contest of the linux magazin
* (www.linux-magazin.de), announced in issue 09/10. It's purpose is to produce
* a TCL-script that connects to a server and play a certain game with another
* program.
*
* It was written by Andreas Barth, Heidenheim/Brenz
*
************************************************************************************
*                                 L I C E N S E
************************************************************************************
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*
************************************************************************************
*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifndef GAME_SERVER
#	define GAME_SERVER "wettbewerb.linux-magazin.de"
#endif
#ifndef GAME_PORT
#	define GAME_PORT 3333
#endif
#ifndef GAME_MAX
#	define GAME_MAX 50  /* the goal is to reach 50 points */
#endif
#ifndef MY_NAME
#	define MY_NAME "asinusphi"
#endif
#define IDX_FACTOR ((GAME_MAX*(GAME_MAX+1))/2)
#define GAME_SIZE (GAME_MAX*GAME_MAX*(GAME_MAX+1)*(GAME_MAX+1)/4) /* number of states that do not mean end of game */
#ifndef ITER_LIMIT
#	define ITER_LIMIT 0.5 /* iteration limit */
#endif
#define NUM_NEXT 7 /* the dice has 6 numbers and we can not chose not to roll the dice -> 7 moves */
#define START_PB 0.5
#define MKINDEX(s) ((s).yN>=GAME_MAX||(s).mN>=GAME_MAX)?GAME_SIZE:(((((int)((s).mN))*((int)((s).mN+1)))/2+(s).mS)*IDX_FACTOR+(((int)((s).yN))*((int)((s).yN+1)))/2+(s).yS)




/* type to hold a game state */
typedef struct { \
	unsigned char yS; /* your safety belt value */ \
	unsigned char yN; /* your points */ \
	unsigned char mS; /* my safety belt values */ \
	unsigned char mN; /* my points */ \
	double chanceM; /* my chance under the assumption that it is my turn */ \
	double chanceY; /* your chance under the assumption that it is our turn */ \
} cState; 

/* array to store all possible states of a game */
static cState states[GAME_SIZE+1]; 
static int numIterations=0;

/* writing the body of the deciding proc in the tcl-script */
static int oldIDX=0, oldhl=0, summe=0;
void shipout(int idx, int hl) {
	if(hl==1 && oldhl==0) {
		oldIDX = idx;
		oldhl  = 1;
	} else if(hl==0 && oldhl==1) {
		if(oldIDX==(idx-1)) { /* only one index */
			printf("\tif {$index == %d} {return 1}\n",oldIDX);
		} else {
			printf("\tif {$index >= %d && $index <= %d} {return 1}\n",oldIDX, idx-1);
		}
		oldhl  = 0;
	}
}

/* writeout the beginning of the tcl-script */
void writeIntro() {
	printf("# some configurations:\n");
	printf("set GAME_SERVER %s\n",GAME_SERVER);
	printf("set GAME_PORT %d\n",GAME_PORT);
	printf("\n");
	printf("set CMD_HELO HELO\n");
	printf("set CMD_AUTH AUTH\n");
	printf("set CMD_DENY DENY\n");
	printf("\n");
	printf("set CMD_TURN TURN\n");
	printf("set CMD_THRW THRW\n");
	printf("set CMD_DEF DEF\n");
	printf("set CMD_WIN WIN\n");
	printf("\n");
	printf("set CMD_ROLL ROLL\n");
	printf("set CMD_SAVE SAVE\n");
	printf("\n");
	printf("set STATE_ROLL  2\n");
	printf("set STATE_YOURS 4\n");
	printf("set STATE_MINE  8\n");
	printf("\n");
	printf("set MY_NAME %s\n",MY_NAME);
	printf("\n");
	printf("# we need some bookkeeping\n");
	printf("# mN: my points  mS: my saved points  yN: opponents points  yS: opponents saved points\n");
	printf("set mN 0\n");
	printf("set mS 0\n");
	printf("set yN 0\n");
	printf("set yS 0\n");
	printf("set state $STATE_YOURS\n");
	printf("\n");
	printf("\n");
	printf("# decide wheather to roll the dice or save the points\n");
	printf("proc decide {mN mS yN yS} {\n");
	printf("\tset index [expr (($mN*($mN+1))/2+$mS)*%d+($yN*($yN+1))/2+$yS]\n",IDX_FACTOR);
	printf("\tif {$index > %d} {set index %d}\n",GAME_SIZE,GAME_SIZE);
}

/* writeout the rest ofthe tcl-script */
void writeRest() {
	printf("\treturn 0\n");
	printf("}\n");
	printf("\n");
	printf("# On first call of a proc, the bytecode is compiled. So we make a dummy call with respect to the time limit:\n");
	printf("decide 0 0 0 0\n\n");
	printf("proc putsf {a b} {\n");
	printf("	puts $a $b\n");
	printf("	flush $a\n");
	printf("}\n");
	printf("\n");
	printf("\n");
	printf("# open network connection\n");
	printf("set io [socket $GAME_SERVER $GAME_PORT]\n");
	printf("set iog $io\n");
	printf("set iop $io\n");
	printf("\n");
	printf("# Now there should be a HELO message for us:\n");
	printf("set line [gets $iog]\n");
	printf("set command [split $line]\n");
	printf("\n");
	printf("if {[string compare -nocase [lindex $command 0] $CMD_HELO] != 0} {\n");
	printf("	# there is no sense in long error messages\n");
	printf("	exit -1\n");
	printf("}\n");
	printf("\n");
	printf("# We got the HELO now we send the AUTH:\n");
	printf("putsf $iop \"$CMD_AUTH $MY_NAME\"\n");
	printf("\n");
	printf("\n");
	printf("# So much for that, now the real thing...\n");
	printf("while {1>0} {\n");
	printf("	set command [split [gets $iog]]\n");
	printf("	set cmd [lindex $command 0]\n");
	printf("\n");
	printf("	if {[string compare -nocase $cmd $CMD_DENY]==0} {\n");
	printf("		# we are kicked off :-(\n");
	printf("		exit -2\n");
	printf("	} elseif {[string compare -nocase $cmd $CMD_TURN]==0} {\n");
	printf("		if {$state==$STATE_YOURS} {\n");
	printf("			# You must have saved your points\n");
	printf("			set yS $yN\n");
	printf("		}\n");
	printf("		# it is our turn so we need to make a decision\n");
	printf("		if {[decide $mN $mS $yN $yS] == 1} {\n");
	printf("			# We save\n");
	printf("			set mS $mN\n");
	printf("			putsf $iop \"$CMD_SAVE\"\n");
	printf("			set state $STATE_YOURS\n");
	printf("		} else {\n");
	printf("			# We roll the dice\n");
	printf("			putsf $iop \"$CMD_ROLL\"\n");
	printf("			set state $STATE_ROLL\n");
	printf("		}\n");
	printf("	} elseif {[string compare -nocase $cmd $CMD_THRW]==0} {\n");
	printf("		set points [lindex $command 1]\n");
	printf("		if {$state==$STATE_ROLL} {\n");
	printf("			# The dice rolled for us.\n");
	printf("			if {$points == 6} {\n");
	printf("				# we go back to the saved poits\n");
	printf("				set mN $mS\n");
	printf("				# now it is your turn:\n");
	printf("				set state $STATE_YOURS\n");
	printf("			} else {\n");
	printf("				# ok, we gain some points\n");
	printf("				incr mN $points\n");
	printf("				set state $STATE_MINE\n");
	printf("			}\n");
	printf("		} else {\n");
	printf("			# The dice was rolled by you.\n");
	printf("			if {$points == 6} {\n");
	printf("				# I'm so sorry for you...\n");
	printf("				set yN $yS\n");
	printf("				set state $STATE_MINE\n");
	printf("			} else {\n");
	printf("				# it's your harvest...\n");
	printf("				incr yN $points\n");
	printf("				set state $STATE_YOURS\n");
	printf("			}\n");
	printf("		}\n");
	printf("	} elseif {[string compare -nocase $cmd $CMD_WIN]==0} {\n");
	printf("		# WE WON! ;-)\n");
	printf("		puts stdout \"We won!\"\n");
	printf("		exit 0\n");
	printf("	} elseif {[string compare -nocase $cmd $CMD_DEF]==0} {\n");
	printf("		# WE LOST! 8-[\n");
	printf("		puts stdout \"We lost!\"\n");
	printf("		exit 0\n");
	printf("	}\n");
	/* printf("	puts stdout \"	Spielstand:  <$mN $mS   $yN $yS> state: $state\"\n"); */
	printf("}\n");
}

/* get the 7 possible next states if it is my turn */
void getMyNext(cState s, int *r) {
	cState d;

	d=s; d.mS=s.mN; r[0]=MKINDEX(d);
	d=s; d.mN+=1; r[1]=MKINDEX(d);
	d=s; d.mN+=2; r[2]=MKINDEX(d);
	d=s; d.mN+=3; r[3]=MKINDEX(d);
	d=s; d.mN+=4; r[4]=MKINDEX(d);
	d=s; d.mN+=5; r[5]=MKINDEX(d);
	d=s; d.mN=s.mS; r[6]=MKINDEX(d);
}

/* get the 7 possible next states if it is your turn */
void getYourNext(cState s, int *r) {
	cState d;

	d=s; d.yS=s.yN; r[0]=MKINDEX(d);
	d=s; d.yN+=1; r[1]=MKINDEX(d);
	d=s; d.yN+=2; r[2]=MKINDEX(d);
	d=s; d.yN+=3; r[3]=MKINDEX(d);
	d=s; d.yN+=4; r[4]=MKINDEX(d);
	d=s; d.yN+=5; r[5]=MKINDEX(d);
	d=s; d.yN=s.yS; r[6]=MKINDEX(d);
}

int main(char *args[], int argc) {
	cState actual;
	unsigned char iyS, iyN, imS, imN;
	int idx;
	double enhancement;

	/* setup the states */
	for(imN=0; imN<GAME_MAX; imN++) {
		for(imS=0; imS<=imN; imS++) {
			for(iyN=0; iyN<GAME_MAX; iyN++) {
				for(iyS=0; iyS<=iyN; iyS++) {
					actual.yS=iyS;
					actual.mS=imS;
					actual.yN=iyN;
					actual.mN=imN;
					actual.chanceM=START_PB;
					actual.chanceY=START_PB;
					states[MKINDEX(actual)]=actual;
				}
			}
		}
	}

	/* now we add the won-state */
	actual.yS=0;
	actual.mS=0;
	actual.yN=GAME_MAX;
	actual.mN=GAME_MAX;
	actual.chanceM=1.0;
	actual.chanceY=1.0;
	states[GAME_SIZE]=actual;

	/* Now the chances: we iterate until they are (nearly) consistent */
	enhancement = ITER_LIMIT+1.0;
	while(enhancement > ITER_LIMIT) {
		enhancement=0.0;
		for(idx=GAME_SIZE-1; idx>=0; idx--) {
			int mNext[NUM_NEXT];
			int yNext[NUM_NEXT];

			double ncM, ocM, c0M, c1M, c2M, c3M, c4M, c5M, c6M;
			double ncY, ocY, c0Y, c1Y, c2Y, c3Y, c4Y, c5Y, c6Y;

			getMyNext(states[idx],mNext);
			getYourNext(states[idx],yNext);

			ocM=states[idx].chanceM;
			ocY=states[idx].chanceY;

			c0M=1.0-states[mNext[0]].chanceY;
			c1M=states[mNext[1]].chanceM;
			c2M=states[mNext[2]].chanceM;
			c3M=states[mNext[3]].chanceM;
			c4M=states[mNext[4]].chanceM;
			c5M=states[mNext[5]].chanceM;
			c6M=1.0-states[mNext[6]].chanceY;

			c0Y=1.0-states[yNext[0]].chanceM;
			c1Y=states[yNext[1]].chanceY;
			c2Y=states[yNext[2]].chanceY;
			c3Y=states[yNext[3]].chanceY;
			c4Y=states[yNext[4]].chanceY;
			c5Y=states[yNext[5]].chanceY;
			c6Y=1.0-states[yNext[6]].chanceM;

			ncM = (c1M+c2M+c3M+c4M+c5M+c6M)/6.0;
			ncM = c0M>ncM?c0M:ncM;
			enhancement += sqrt((ncM-ocM)*(ncM-ocM));
			states[idx].chanceM = ncM;

			ncY = (c1Y+c2Y+c3Y+c4Y+c5Y+c6Y)/6.0;
			ncY = c0Y>ncY?c0Y:ncY;
			enhancement += sqrt((ncY-ocY)*(ncY-ocY));
			states[idx].chanceY = ncY;
		}
		numIterations++;
	}

	/* just for information purposes: */
	fprintf(stderr,"Needed %d iterations for precision %g (%g)\n",numIterations,enhancement,ITER_LIMIT);

	/* Now we know everything and cat write the tcl-script tha does the real job. */
	writeIntro();

	for(idx=0; idx<=GAME_SIZE; idx++) {
		cState s=states[idx];
		int n[7];
		double cM=s.chanceM;
		double cY=s.chanceY;
		double cR;
		double cS;

		getMyNext(s,n);
		cR=(states[n[1]].chanceM+states[n[2]].chanceM+states[n[3]].chanceM+states[n[4]].chanceM+
			states[n[5]].chanceM+(1-states[n[6]].chanceY))/6.0;
		cS = 1.0-states[n[0]].chanceY;

		if(cR<cS){
			shipout(idx,1);
		} else {
			shipout(idx,0);
		}
	}
	writeRest();
	exit(0);
}
